#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/spi/spi.h>
#include <linux/gpio.h>
#include <linux/interrupt.h>
#include <linux/delay.h>

#define GPIO_IRQ 1992          // Replace with your GPIO number
#define SPI_BUS_SPEED 500000   // SPI speed in Hz
#define SPI_BITS_PER_WORD 8    // SPI bits per word
#define SPI_MODE SPI_MODE_0    // SPI mode

static struct spi_device *spi_dev;
static int gpio_irq;
static unsigned char tx_buffer[] = "Hello";
static unsigned char rx_buffer[sizeof(tx_buffer)];

static int perform_spi_transfer(void)
{
    struct spi_transfer transfer = {
        .tx_buf = tx_buffer,
        .rx_buf = rx_buffer,
        .len = sizeof(tx_buffer),
        .speed_hz = SPI_BUS_SPEED,
        .bits_per_word = SPI_BITS_PER_WORD,
    };
    struct spi_message msg;
    int ret;

    spi_message_init(&msg);
    spi_message_add_tail(&transfer, &msg);

    printk(KERN_INFO "Performing SPI transfer\n");
    ret = spi_sync(spi_dev, &msg);
    if (ret < 0) {
        printk(KERN_ERR "SPI transfer failed: %d\n", ret);
        return ret;
    }

    printk(KERN_INFO "SPI transfer successful. Received data: %s\n", rx_buffer);
    return 0;
}

static irqreturn_t gpio_irq_handler(int irq, void *dev_id)
{
    printk(KERN_INFO "GPIO interrupt triggered\n");

    // Perform SPI transfer
    perform_spi_transfer();

    return IRQ_HANDLED;
}

static int my_spi_probe(struct spi_device *spi)
{
    spi_dev = spi;

    // Configure SPI settings
    spi->mode = SPI_MODE;
    spi->max_speed_hz = SPI_BUS_SPEED;
    spi->bits_per_word = SPI_BITS_PER_WORD;
    if (spi_setup(spi)) {
        printk(KERN_ERR "Failed to setup SPI device\n");
        return -ENODEV;
    }

    printk(KERN_INFO "SPI device probed: %s\n", dev_name(&spi->dev));
    return 0;
}

static void my_spi_remove(struct spi_device *spi)
{
    spi_dev = NULL;
    printk(KERN_INFO "SPI device removed\n");
}

static const struct of_device_id my_spi_of_match[] = {
    { .compatible = "my,spi-device", },
    { /* sentinel */ }
};
MODULE_DEVICE_TABLE(of, my_spi_of_match);

static const struct spi_device_id my_spi_id[] = {
    { "my,spi-device", 0 },
    { /* sentinel */ }
};
MODULE_DEVICE_TABLE(spi, my_spi_id);

static struct spi_driver my_spi_driver = {
    .driver = {
        .name = "my_spi_driver",
        .of_match_table = my_spi_of_match,
        .owner = THIS_MODULE,
    },
    .probe = my_spi_probe,
    .remove = my_spi_remove,
    .id_table = my_spi_id,
};

static int __init my_spi_gpio_init(void)
{
    int ret;

    printk(KERN_INFO "Initializing SPI GPIO module\n");

    // Register SPI driver
    ret = spi_register_driver(&my_spi_driver);
    if (ret < 0) {
        printk(KERN_ERR "Failed to register SPI driver: %d\n", ret);
        return ret;
    }

    // Request GPIO
    if (!gpio_is_valid(GPIO_IRQ)) {
        printk(KERN_ERR "Invalid GPIO\n");
        spi_unregister_driver(&my_spi_driver);
        return -ENODEV;
    }

    ret = gpio_request(GPIO_IRQ, "my_gpio");
    if (ret) {
        printk(KERN_ERR "Failed to request GPIO: %d\n", ret);
        spi_unregister_driver(&my_spi_driver);
        return ret;
    }

    // Configure GPIO as input
    ret = gpio_direction_input(GPIO_IRQ);
    if (ret) {
        printk(KERN_ERR "Failed to set GPIO direction: %d\n", ret);
        gpio_free(GPIO_IRQ);
        spi_unregister_driver(&my_spi_driver);
        return ret;
    }

    // Request interrupt
    gpio_irq = gpio_to_irq(GPIO_IRQ);
    ret = request_irq(gpio_irq, gpio_irq_handler, IRQF_TRIGGER_RISING, "my_gpio_irq", NULL);
    if (ret) {
        printk(KERN_ERR "Failed to request IRQ: %d\n", ret);
        gpio_free(GPIO_IRQ);
        spi_unregister_driver(&my_spi_driver);
        return ret;
    }

    printk(KERN_INFO "SPI GPIO module initialized\n");
    return 0;
}

static void __exit my_spi_gpio_exit(void)
{
    printk(KERN_INFO "Exiting SPI GPIO module\n");

    free_irq(gpio_irq, NULL);
    gpio_free(GPIO_IRQ);
    spi_unregister_driver(&my_spi_driver);

    printk(KERN_INFO "SPI GPIO module exited\n");
}

module_init(my_spi_gpio_init);
module_exit(my_spi_gpio_exit);

MODULE_LICENSE("GPL");
MODULE_DESCRIPTION("SPI GPIO Interrupt Example");
MODULE_AUTHOR("Your Name");
